@@ -1,9 +1,19 @@
+"""
+TSX Functions
+---------------
+
+Set of functions to scrap ticker data from the toronto stock exchange.
+
+Will definitely split into smaller files once the graphql api becomes the main api.
+"""
+
+
 import requests
 import pandas as pd
 import multiprocessing as mp
+from concurrent.futures import ThreadPoolExecutor
 from json import loads
 from bs4 import BeautifulSoup
-from datetime import datetime
 
 # importing the sys module 
 import sys 
@@ -17,7 +27,10 @@ import sys
 sys.setrecursionlimit(10**6) 
 
 
-def get_description_for_ticker(ticker): # change the body of the loop to function
+def get_description_for_ticker(ticker: str): # change the body of the loop to function
+  """
+    set of functionality
+  """
   # grab ticker return desc
   # do on google colab
   # https://stackoverflow.com/questions/56987872/parallelize-pandas-column-update
@@ -26,16 +39,22 @@ def get_description_for_ticker(ticker): # change the body of the loop to functio
     return ''
   return company_description_by_ticker(symbol)
 
-def get_mig_report(filename='', exchange="TSX", return_df=False) -> str:
-  """Gets excel spreadsheet from api.tsx using requests
-  Input
+def get_mig_report(filename: str='', exchange: str="TSX", return_df: bool =False) -> str:
+  """
+  Description:
+    Gets excel spreadsheet from tsx api programatically.
+    See for more flexibility :func:`dl_tsx_xlsx <cad_tickers.exchanges.tsx.dl_tsx_xlsx>`
+
+  Parameters:
     filename: Name of the file to be saved
     exchanges: TSX, TSXV
     return_df: Return a pandas dataframe
-  Output:
-    filePath returns path to file
+
+  Returns:
+    filePath: returns path to file or dataframe
 
   See ://stackoverflow.com/questions/13567507/passing-csrftoken-with-python-requests
+
   """
   tsx_url = 'https://api.tmxmoney.com/en/migreport/search'
   client = requests.session()
@@ -50,9 +69,7 @@ def get_mig_report(filename='', exchange="TSX", return_df=False) -> str:
   login_data = dict(report_type='excel', csrfmiddlewaretoken=csrftoken)
   if exchange in ["TSX", "TSXV"]:
     login_data["exchanges"] = exchange
-  # To get all exchanges, leave blank
-  # login_data = dict(report_type='excel', csrfmiddlewaretoken=client.cookies['csrftoken'])
-  # Options are TSX, TSXV and nothing for all
+
   r = client.post("https://api.tmxmoney.com/en/migreport/search", data=login_data, headers=dict(Referer=tsx_url))
   resp_data = r.content
   if return_df is True:
@@ -69,19 +86,19 @@ def get_mig_report(filename='', exchange="TSX", return_df=False) -> str:
     else:
       return None
 
-# exchanges: TSXV
-# marketcap: 
-# hqregions: canada
-# hqlocations: 
-# sectors:
-# //*[@id="root"]/div[4]/div[3]/div[3]/div[1]/div[2]/div[1]/div[2]/div[1]
-def grab_symbol_for_ticker(ticker) -> str:
-  """Grabs the first symbol from ticker data
-
-  all symbols should lead to valid webpages for data scrapping
-
-  TODO: If anyone wants to validate that, be my guest
+def grab_symbol_for_ticker(ticker: str) -> str:
   """
+  Description:
+    Grabs the first symbol from ticker data
+    all symbols should lead to valid webpages for data scrapping.
+
+  Parameters: 
+    ticker: string representing the stock ticker
+
+  Returns: 
+    symbol: string - searchable string in the quotemedia api or empty string
+  """
+
   if ticker is None or ticker is '':
     return ''
 
@@ -98,17 +115,52 @@ def grab_symbol_for_ticker(ticker) -> str:
       return ''
   return ticker_data[0].get('symbol')
 
-def add_descriptions_to_df_pp(df) -> pd.DataFrame:
+
+def add_descriptions_to_df_pp(df: pd.DataFrame, max_workers: int = 16) -> pd.DataFrame:
+  """
+  Description: fetch descriptions for tickers in parallel
+  noticable speedup uses thread pool which should be faster
+
+  Input:
+    df: dataframe containing tickers
+  Returns:
+    df: updated dataframe with a descriptions if available
+  """
+  df['description'] = ''
+  tickers = df['Ticker'].tolist()
+  with ThreadPoolExecutor(max_workers=max_workers) as tpe:
+    iterables = tpe.map(get_description_for_ticker, tickers)
+    descriptions = list(iterables)
+  df['description'] = descriptions
+  return df
+
+def add_descriptions_to_df_pp_legacy(df: pd.DataFrame) -> pd.DataFrame:
+  """
+  Description: fetch descriptions for tickers in parallel
+  noticable speedup, keeping this to verify speed increase
+
+  Input:
+    df: dataframe containing tickers
+  Returns:
+    df: updated dataframe with a descriptions if available
+  """
   df['description'] = ''
   tickers = df['Ticker'].tolist()
   with mp.Pool() as p:
     descriptions = p.map(get_description_for_ticker, tickers)
   df['description'] = descriptions
-  df.to_csv('tsx_descriptions.csv')
   return df
 
 # https://stackoverflow.com/questions/56987872/parallelize-pandas-column-update
 def add_descriptions_to_df(df) -> pd.DataFrame:
+  """
+  Description: single process solution to fetching descriptions
+
+  Input:
+    df: dataframe containing tickers
+  Returns:
+    df: updated dataframe with a descriptions if available
+  """
   df['description'] = ''
   for index, row in df.iterrows():
     ticker = df.at[index, 'Ticker']
@@ -120,10 +172,18 @@ def add_descriptions_to_df(df) -> pd.DataFrame:
       pass
     description = company_description_by_ticker(symbol)
     df.at[index,'description'] = description
-  df.to_csv('tsx_descriptions.csv')
   return df
 
 def company_description_by_ticker(ticker)-> str:
+  """
+  
+  Description: Grabs searchable ticker from quotemedia using tmx ticker
+
+  Input:
+    ticker: string 
+  Returns:
+    df:  updated dataframe with a descriptions if available
+  """
   # get lookup symbol
   search_symbol = grab_symbol_for_ticker(ticker)
   if search_symbol is '':
@@ -151,21 +211,25 @@ def company_description_by_ticker(ticker)-> str:
     return ''
 
 def lookup_symbol_by_ticker(ticker)-> list:
-  """Returns search dictionary for ticker/
-  sometimes the name of the ticker in the xlsx 
-  sheet is off slightly and we need to find the "real ticker"
+  """
+  
+  Description: Returns search array dictionary for tickers
+  
+  .. note::
 
-  Uses standard api to grab tickers
-  TODO: make another function that uses the new graphql api instead
-  in case this one gets depreciated
-  Example searchpoint is https://app.quotemedia.com/lookup?callback=html&q=zmd&limit=5&webmasterId=101020
+    sometimes the name of the ticker in the xlsx 
+    sheet is off slightly and we need to find the "real ticker".
+    Uses standard api (not graphql) to grab tickers
 
-  See https://app-money.tmx.com/graphql and https://money.tmx.com/en/
-    Ex: {"operationName":"findCompaniesByKeywords","variables":{"keywords":"zmd"},"query":"query findCompaniesByKeywords($keywords: String) {\n  findCompaniesByKeywords(keywords: $keywords) {\n    symbol\n    name\n    exchange\n    __typename\n  }\n}\n"}
+    Example searchpoint is https://app.quotemedia.com/lookup?callback=tmxtickers&q=zmd&limit=5&webmasterId=101020
 
-  Input: Ticker as string
+  See `Tmx Graphql <https://app-money.tmx.com/graphql>`_  and the new `tmx site <https://money.tmx.com/en/>`_
 
-  Output: Dictionary
+  Input:
+    ticker: tmx ticker
+
+  Output:
+    quote_data: list of ticker metadata
   """
   if ticker is None:
     return []
@@ -182,7 +246,7 @@ def lookup_symbol_by_ticker(ticker)-> list:
   quote_data = r.text
 
   if quote_data is None:
-    print('Data missing, no response available')
+     raise Exception('No data returned, check if api is depreciated or contact author for help.')
   s_idx = quote_data.find('(')
   e_idx = quote_data.rfind(')')
 
@@ -199,22 +263,29 @@ def lookup_symbol_by_ticker(ticker)-> list:
     exit(1)
 
 def dl_tsx_xlsx(filename = None, **kwargs) -> str:
-  """Gets excel spreadsheet from api.tsx using requests
-  
-  Replicates api calls in TSX discover tool
-
-  Input
-    filename: Name of the file to be saved
-    options: 
-      exchanges: TSX, TSXV
-      marketcap: values from 0 to specified value
-      sectors: cpc, clean-technology, closed-end-funds, technology
-      # Other options depending on what is selected
-  Output:
-    filePath: returns path to file or pandas dataframe
-
-  See ://stackoverflow.com/questions/13567507/passing-csrftoken-with-python-requests
   """
+  Description: Gets excel spreadsheet from the tsx api using programatically
+
+  .. note::
+
+    Replicates api calls in TSX discover tool with all parameters.
+    See `migreport search <https://api.tmxmoney.com/en/migreport/search>`_
+    Note that not all parameters are documented and/or limited validation
+
+  Parameters:
+    filename: Name of the file to be saved
+
+  Kwargs:
+    * exchanges (string): TSX, TSXV
+    * marketcap (string): values from 0 to specified value
+    * sectors (string): cpc, clean-technology, closed-end-funds, technology
+
+  Returns:
+    data: returns path to file or pandas dataframe
+
+  See `passing csrftoken <https://stackoverflow.com/questions/13567507/passing-csrftoken-with-python-requests>`_
+  """
+
   tsx_url = 'https://api.tmxmoney.com/en/migreport/search'
   client = requests.session()
   # Retrieve the CSRF token first
@@ -237,9 +308,6 @@ def dl_tsx_xlsx(filename = None, **kwargs) -> str:
     else:
       search_data[k] = v
 
-  # To get all exchanges, leave blank
-  # login_data = dict(report_type='excel', csrfmiddlewaretoken=client.cookies['csrftoken'])
-  # Options are TSX, TSXV and nothing for all
   r = client.post(tsx_url, data=search_data, headers=dict(Referer=tsx_url))
   resp_headers = r.headers
   if "text/html" in resp_headers["Content-Type"]:
@@ -249,14 +317,15 @@ def dl_tsx_xlsx(filename = None, **kwargs) -> str:
     if filename is None:
       return pd.read_excel(resp_data)
     else:
-      with open(filename, 'wb') as s:
-        s.write(resp_data)
+      with open(filename, 'wb') as f_:
+        f_.write(resp_data)
       return filename
   else:
     return None
 
 # Add endpoint to query search parameters for tsx spreadsheets
 if __name__ == "__main__":
+  from datetime import datetime
   start_time = datetime.now()
   import argparse
   parser = argparse.ArgumentParser()
@@ -273,11 +342,7 @@ if __name__ == "__main__":
                     help='TSX, TSXV, All (default: %(default)s)') 
   args = parser.parse_args()
   df = dl_tsx_xlsx(sectors='technology')
-  # df = add_descriptions_to_df(df)
   df = add_descriptions_to_df_pp(df)
-  # print(df)
-  len(df)
+  print(df)
   end_time = datetime.now()
   print(end_time - start_time)
-  # get_mig_report(args.file, args.exchange)
-  # company_description_by_ticker('zmd')
